@@ -1,6 +1,7 @@
 """
 
 """
+import socket
 import os
 import io
 import numpy as np
@@ -13,15 +14,20 @@ import boto3
 import botocore
 import smtplib
 import ssl
+import requests
 from pandas.core.groupby import SeriesGroupBy, GroupBy
 import orjson
 from time import sleep
 import traceback
 from hashlib import blake2b
-from tethys_utils.data_models import Geometry, Dataset, DatasetBase, S3ObjectKey, Station, Stats
+from tethys_utils.data_models import Geometry, Dataset, DatasetBase, S3ObjectKey, Station, Stats, StationBase
+# from data_models import Geometry, Dataset, DatasetBase, S3ObjectKey, Station, Stats, StationBase
 from geojson import Point
 import urllib3
-from multiprocessing.pool import ThreadPool
+from multiprocessing.pool import ThreadPool, Pool
+from tethysts.utils import key_patterns, get_object_s3
+from email.message import EmailMessage
+
 
 ####################################################
 ### Misc reference objects
@@ -34,13 +40,6 @@ nc_ts_key_pattern = {
                     'H25': 'tethys/diff/{dataset_id}/{date}.H25.nc.zst'
                     }
 
-key_patterns = {'results': 'tethys/v2/{dataset_id}/{station_id}/{run_date}/results.nc.zst',
-                'datasets': 'tethys/v2/datasets.json.zst',
-                'stations': 'tethys/v2/{dataset_id}/stations.json.zst',
-                'station': 'tethys/v2/{dataset_id}/{station_id}/station.json.zst',
-                'dataset': 'tethys/v2/{dataset_id}/dataset.json.zst',
-                }
-
 base_ds_fields = ['feature', 'parameter', 'method', 'product_code', 'owner', 'aggregation_statistic', 'frequency_interval', 'utc_offset']
 
 ####################################################
@@ -87,7 +86,26 @@ def s3_connection(conn_config, max_pool_connections=20):
 
 def list_parse_s3(s3_client, bucket, prefix, start_after='', delimiter='', continuation_token=''):
     """
+    Wrapper S3 function around the list_objects_v2 base function with a Pandas DataFrame output.
 
+    Parameters
+    ----------
+    s3_client : boto3.client
+        A boto3 client object
+    bucket : str
+        The S3 bucket.
+    prefix : str
+        Limits the response to keys that begin with the specified prefix.
+    start_after : str
+        The S3 key to start after.
+    delimiter : str
+        A delimiter is a character you use to group keys.
+    continuation_token : str
+        ContinuationToken indicates to S3 that the list is being continued on this bucket with a token.
+
+    Returns
+    -------
+    DataFrame
     """
     if s3_client._endpoint.host == 'https://vault.revera.co.nz':
         js = []
@@ -133,12 +151,12 @@ def list_parse_s3(s3_client, bucket, prefix, start_after='', delimiter='', conti
     return f_df1
 
 
-def get_last_date(s3_df, default_date='1900-01-01', date_type='date', local_tz=None):
+def get_last_date(s3_df, default_date='1900-01-01', date_type='date'):
     """
 
     """
     if not s3_df.empty:
-        last_run_date = s3_df['KeyDate'].max().tz_convert(local_tz).tz_localize(None)
+        last_run_date = s3_df['KeyDate'].max()
     else:
         last_run_date = pd.Timestamp(default_date)
 
@@ -288,11 +306,12 @@ def results_data_integrety_checks(data, param_name, attrs, encoding, ancillary_v
     # Time series data
     data_cols = []
     if isinstance(ancillary_variables, list):
-        for av in ancillary_variables:
-            if not av in data:
-                raise ValueError('The DataFrame must contain every value in the ancillary_variables list')
-            else:
-                data_cols.extend([av])
+        if len(ancillary_variables) > 0:
+            for av in ancillary_variables:
+                if not av in data:
+                    raise ValueError('The DataFrame must contain every value in the ancillary_variables list')
+                else:
+                    data_cols.extend([av])
 
     data_cols.extend([param_name])
 
@@ -331,21 +350,21 @@ def results_data_integrety_checks(data, param_name, attrs, encoding, ancillary_v
     return data
 
 
-def station_data_integrety_checks(data, attrs=None, encoding=None):
-    """
+# def station_data_integrety_checks(data, attrs=None, encoding=None):
+#     """
 
-    """
-    # Station data
-    essentials = {'lat': (float, np.float), 'lon': (float, np.float), 'station_id': str, 'altitude': (int, float, np.float, np.int)}
+#     """
+#     # Station data
+#     essentials = {'lat': (float, np.float), 'lon': (float, np.float), 'station_id': str, 'altitude': (int, float, np.float, np.int)}
 
-    for c in essentials:
-        if not c in data.keys():
-            raise ValueError('The station_data DataFrame must contain the field: ' + str(c))
+#     for c in essentials:
+#         if not c in data.keys():
+#             raise ValueError('The station_data DataFrame must contain the field: ' + str(c))
 
-        if not isinstance(data[c], essentials[c]):
-            raise ValueError('The station_data DataFrame field must have the data type(s): ' + str(essentials[c]))
+#         if not isinstance(data[c], essentials[c]):
+#             raise ValueError('The station_data DataFrame field must have the data type(s): ' + str(essentials[c]))
 
-    return data
+#     return data
 
 
 def data_to_xarray(results_data, station_data, param_name, results_attrs, results_encoding, station_attrs=None, station_encoding=None, virtual_station=False, run_date=None, ancillary_variables=None, compression=False, compress_level=1):
@@ -381,7 +400,11 @@ def data_to_xarray(results_data, station_data, param_name, results_attrs, result
     ts_data1 = results_data_integrety_checks(results_data, param_name, results_attrs, results_encoding, ancillary_variables)
 
     # Station data
-    station_data1 = station_data_integrety_checks(station_data)
+    stn_m = StationBase(**station_data)
+    stn_data = orjson.loads(stn_m.json(exclude_none=True))
+    stn_data['lon'] = stn_data['geometry']['coordinates'][0]
+    stn_data['lat'] = stn_data['geometry']['coordinates'][1]
+    stn_data.pop('geometry')
 
     ## Assign Attributes
 
@@ -391,9 +414,9 @@ def data_to_xarray(results_data, station_data, param_name, results_attrs, result
         attrs1 = {}
 
     attrs1.update({'station_id': {'cf_role': "timeseries_id", 'virtual_station': virtual_station}, 'lat': {'standard_name': "latitude", 'units': "degrees_north"}, 'lon': {'standard_name': "longitude", 'units': "degrees_east"}, 'altitude': {'standard_name': 'surface_altitude', 'long_name': 'height above the geoid to the lower boundary of the atmosphere', 'units': 'm'}})
-    if 'name' in station_data.keys():
+    if 'name' in stn_data.keys():
         attrs1.update({'name': {'long_name': 'station name'}})
-    if 'ref' in station_data.keys():
+    if 'ref' in stn_data.keys():
         attrs1.update({'ref': {'long_name': 'station reference id given by the owner'}})
 
     ts_cols = list(results_data.columns)
@@ -406,7 +429,8 @@ def data_to_xarray(results_data, station_data, param_name, results_attrs, result
         attrs1.update({'modified_date': {'long_name': 'last modified date'}})
 
     if isinstance(ancillary_variables, list):
-        attrs1[param_name].update({'ancillary_variables': ' '.join(ancillary_variables)})
+        if len(ancillary_variables) > 0:
+            attrs1[param_name].update({'ancillary_variables': ' '.join(ancillary_variables)})
 
     ## Assign encodings
     if isinstance(station_encoding, dict):
@@ -434,7 +458,7 @@ def data_to_xarray(results_data, station_data, param_name, results_attrs, result
     ## Create the Xarray Dataset
 
     ds1 = results_data.to_xarray()
-    for k, v in station_data.items():
+    for k, v in stn_data.items():
         ds1[k] = v
 
     ## Add attributes and encodings
@@ -543,7 +567,7 @@ def grp_ts_agg(df, grp_col, ts_col, freq_code, agg_fun, discrete=False, **kwargs
         raise TypeError('The object must be a DataFrame')
 
 
-def compare_dfs(old_df, new_df, on, parameter):
+def compare_dfs(old_df, new_df, on, parameter, add_old=False):
     """
     Function to compare two DataFrames with nans and return a dict with rows that have changed (diff), rows that exist in new_df but not in old_df (new), and rows  that exist in old_df but not in new_df (remove).
     Both DataFrame must have the same columns. If both DataFrames are identical, and empty DataFrame will be returned.
@@ -572,7 +596,6 @@ def compare_dfs(old_df, new_df, on, parameter):
 
     comp1 = pd.merge(old_df, new_df, on=on, how='outer', indicator=True, suffixes=('_x', ''))
 
-    # rem1 = comp1.loc[comp1._merge == 'left_only', on].copy()
     add_set = comp1.loc[comp1._merge == 'right_only', all_cols].copy()
     comp2 = comp1[comp1._merge == 'both'].drop('_merge', axis=1).copy()
 
@@ -606,12 +629,20 @@ def compare_dfs(old_df, new_df, on, parameter):
         diff_set = new_set[c2].copy()
         old_set2 = old_set[~c2].copy()
 
-        all_set = pd.concat([old_set2, diff_set, add_set])
+        if add_old:
+            not_cols = list(on)
+            [not_cols.extend([c]) for c in comp1.columns if '_x' in c]
+            add_old1 = comp1.loc[comp1._merge == 'left_only', not_cols].copy()
+            add_old1.rename(columns=old_cols_map, inplace=True)
+
+            all_set = pd.concat([old_set2, diff_set, add_set, add_old1])
+        else:
+            all_set = pd.concat([old_set2, diff_set, add_set])
 
     return all_set
 
 
-def compare_xrs(old_xr, new_xr):
+def compare_xrs(old_xr, new_xr, add_old=False):
     """
 
     """
@@ -642,7 +673,7 @@ def compare_xrs(old_xr, new_xr):
     # old_df['modified_date'] = pd.Timestamp('2020-12-29')
 
     ## run comparison
-    comp = compare_dfs(old_df, new_df, on, parameter)
+    comp = compare_dfs(old_df, new_df, on, parameter, add_old=add_old)
 
     if comp.empty:
         print('Nothing has changed. Returning empty DataFrame.')
@@ -710,7 +741,7 @@ def discrete_resample(df, freq_code, agg_fun, remove_inter=False, **kwargs):
     return s6
 
 
-def email_msg(sender_address, sender_password, receiver_address, subject, body):
+def email_msg(sender_address, sender_password, receiver_address, subject, body, smtp_server="smtp.gmail.com"):
     """
     Function to send a simple email using gmail smtp.
 
@@ -726,6 +757,8 @@ def email_msg(sender_address, sender_password, receiver_address, subject, body):
         The subject of the email.
     body : str
         The main body of the email.
+    smtp_server : str
+        The SMTP server to send the email through.
 
     Returns
     -------
@@ -733,19 +766,38 @@ def email_msg(sender_address, sender_password, receiver_address, subject, body):
 
     """
     port = 465  # For SSL
-    smtp_server = "smtp.gmail.com"
 
-    msg_base = """Subject: {subject}\n
-    {body}"""
+    msg = EmailMessage()
 
-    msg = msg_base.format(subject=subject, body=body)
+    # msg_base = """From: {from1}\n
+    # To: {to}\n
+    # Subject: {subject}\n
+
+    # {body}
+
+    # hostname: {host}
+    # IP address: {ip}"""
+
+    ip = requests.get('https://api.ipify.org').text
+
+    body_base = """{body}
+
+    hostname: {host}
+    IP address: {ip}"""
+
+    # msg = msg_base.format(subject=subject, body=body, host=socket.getfqdn(), ip=ip, from1=sender_address, to=receiver_address)
+
+    msg['From'] = sender_address
+    msg['To'] = receiver_address
+    msg['Subject'] = subject
+    msg.set_content(body_base.format(body=body, host=socket.getfqdn(), ip=ip))
 
     # Create a secure SSL context
     context = ssl.create_default_context()
 
     with smtplib.SMTP_SSL(smtp_server, port, context=context) as server:
         server.login(sender_address, sender_password)
-        server.sendmail(sender_address, receiver_address, msg)
+        server.send_message(msg)
 
 
 def tsreg(ts, freq=None, interp=False):
@@ -800,7 +852,7 @@ def assign_ds_ids(datasets):
     ### Iterate through the dataset list
     for ds in dss:
         # print(ds)
-        ### Validate model
+        ## Validate base model
         ds_m = DatasetBase(**ds)
 
         base_ds = {k: ds[k] for k in base_ds_fields}
@@ -809,9 +861,27 @@ def assign_ds_ids(datasets):
 
         ds['dataset_id'] = ds_id
 
+        ## Validate full model
+        ds_m = Dataset(**ds)
+
     return dss
 
 
+def process_datasets(datasets):
+    """
+
+    """
+    for ht_ds, ds_list in datasets.items():
+        ds_list2 = assign_ds_ids(ds_list)
+        datasets[ht_ds] = ds_list2
+
+    dataset_list = []
+    [dataset_list.extend(ds_list) for ht_ds, ds_list in datasets.items()]
+
+    return dataset_list
+
+
+
 def put_remote_dataset(s3, bucket, dataset, run_date=None):
     """
 
@@ -900,17 +970,20 @@ def assign_station_id(geometry):
     return station_id
 
 
-def ht_stats(df, parameter, precision):
+def get_new_stats(data):
     """
 
     """
-    min1 = df[parameter].min().round(precision)
-    max1 = df[parameter].max().round(precision)
-    from_date = df['time'].min()
-    to_date = df['time'].max()
-    count = df['time'].count()
+    vars1 = list(data.variables)
+    parameter = [v for v in vars1 if 'dataset_id' in data[v].attrs][0]
+    precision = int(np.abs(np.log10(data[parameter].attrs['precision'])))
 
-    stats1 = Stats(min=min1, max=max1, count=count, from_date=from_date, to_date=to_date)
+    min1 = round(float(data[parameter].min()), precision)
+    max1 = round(float(data[parameter].max()), precision)
+    from_date = pd.Timestamp(data['time'].min().values)
+    to_date = pd.Timestamp(data['time'].max().values)
+
+    stats1 = Stats(min=min1, max=max1, from_date=from_date, to_date=to_date)
 
     return stats1
 
@@ -920,30 +993,124 @@ def process_object_keys(s3, bucket, prefix):
 
     """
     keys1 = list_parse_s3(s3, bucket, prefix)
-    keys2 = keys1[keys1.Key.str.contains('results.nc.zst')]
+    keys2 = keys1[keys1.Key.str.contains('results')]
 
     infos1 = [S3ObjectKey(key=row['Key'], bucket=bucket, content_length=row['Size'], etag=row['ETag'], run_date=row['KeyDate'], modified_date=row['LastModified']) for i, row in keys2.iterrows()]
 
     return infos1
 
 
-def process_real_station(dataset_id, coords, df, parameter, precision, s3, bucket, geo_type='Point', ref=None, name=None, osm_id=None, altitude=None, properties=None, mod_date=None):
+def process_stations_df(stns_df):
+    """
+    For processing the initial stations prior to assigning the altitude.
+    """
+    stns1 = stns_df.copy()
+    stns1['coords'] = stns1.apply(lambda x: [x.lon, x.lat], axis=1)
+
+    stns1['geo'] = stns1.apply(lambda x: create_geometry(x.coords), axis=1)
+    stns1['station_id'] = stns1.apply(lambda x: assign_station_id(x.geo), axis=1)
+
+    stns2 = stns1.drop_duplicates('station_id').drop('geo', axis=1).copy()
+
+    return stns2
+
+
+def process_station_base(coords, ref=None, name=None, osm_id=None, altitude=None, properties=None, virtual_station=False, geo_type='Point', return_dict=True):
+    """
+
+    """
+    ## Create geometry and station_id
+    geo = create_geometry(coords, geo_type=geo_type)
+    stn_id = assign_station_id(geo)
+
+    ## Put into data model
+    stn_m = StationBase(station_id=stn_id, geometry=geo, ref=ref, name=name, osm_id=osm_id, altitude=altitude, properties=properties, virtual_station=virtual_station)
+
+    if return_dict:
+        stn_m = orjson.loads(stn_m.json(exclude_none=True))
+
+    return stn_m
+
+
+def process_stations_base(stns_list):
+    """
+
+    """
+    stns_dict = {}
+    for s in stns_list:
+        stn_m = process_station_base(**s)
+        stns_dict[stn_m['station_id']] = stn_m
+
+    return stns_dict
+
+
+def get_station_data_from_xr(data):
+    """
+    Parameters
+    ----------
+    data : xr.Dataset
+    """
+    vars1 = list(data.variables)
+    dims1 = list(data.dims.keys())
+    parameter = [v for v in vars1 if 'dataset_id' in data[v].attrs][0]
+    attrs = data[parameter].attrs.copy()
+    data_vars = [parameter]
+    if 'ancillary_variables' in attrs:
+        ancillary_variables = attrs['ancillary_variables'].split(' ')
+        data_vars.extend(ancillary_variables)
+
+    stn_vars = [v for v in vars1 if (not v in dims1) and (not v in data_vars)]
+    stn_data1 = {k: v['data'] for k, v in data[stn_vars].to_dict()['data_vars'].items()}
+    stn_data1['lon'] = round(stn_data1['lon'], 6)
+    stn_data1['lat'] = round(stn_data1['lat'], 6)
+    stn_data1['altitude'] = round(stn_data1['altitude'], 3)
+
+    return stn_data1
+
+
+def process_station_summ(dataset_id, data, connection_config, bucket, mod_date=None):
     """
 
     """
     if mod_date is None:
         mod_date = pd.Timestamp.today(tz='utc').round('s').tz_localize(None)
+    elif isinstance(mod_date, (str, pd.Timestamp)):
+        mod_date = pd.Timestamp(mod_date).tz_localize(None)
 
-    geo1 = create_geometry(coords, geo_type='Point')
-    station_id = assign_station_id(geo1)
-
-    stats1 = ht_stats(df, parameter, precision)
+    ## Genreate the info for the recently created data
+    station_id = str(data['station_id'].values)
+    stats1 = get_new_stats(data)
 
+    s3 = s3_connection(connection_config)
     prefix = key_patterns['results'].split('{run_date}')[0].format(dataset_id=dataset_id, station_id=station_id)
 
     object_infos1 = process_object_keys(s3, bucket, prefix)
 
-    station_m = Station(dataset_id=dataset_id, station_id=station_id, virtual_station=False, geometry=geo1, stats=stats1, results_object_key=object_infos1, ref=ref, name=name, osm_id=osm_id, altitude=altitude, properties=properties, modified_date=mod_date)
+    ## Get old data
+    stn_key = key_patterns['station'].format(dataset_id=dataset_id, station_id=station_id)
+
+    try:
+        old_stn_data_obj = get_object_s3(stn_key, connection_config, bucket)
+        old_stn_data = read_json_zstd(old_stn_data_obj)
+    except:
+        old_stn_data = {}
+
+    if old_stn_data:
+        old_stats = old_stn_data['stats']
+        min1 = min([old_stats['min'], stats1.min])
+        max1 = max([old_stats['max'], stats1.max])
+        to_date = max([pd.Timestamp(old_stats['to_date']), stats1.to_date])
+        from_date = min([pd.Timestamp(old_stats['from_date']), stats1.from_date])
+
+        stats2 = Stats(min=min1, max=max1, from_date=from_date, to_date=to_date)
+    else:
+        stats2 = stats1
+
+    ## Put it all together
+    stn_dict2 = get_station_data_from_xr(data)
+    stn_dict2.update({'dataset_id': dataset_id, 'stats': stats2, 'results_object_key': object_infos1, 'modified_date': mod_date})
+
+    station_m = Station(**stn_dict2)
 
     return station_m
 
@@ -1058,32 +1225,54 @@ def put_remote_agg_datasets(s3, bucket, threads=20):
     return output
 
 
-def compare_datasets_from_s3(s3, bucket, new_data):
+def compare_datasets_from_s3(conn_config, bucket, new_data, add_old=False, read_buffer=False, last_run_date_key=None, public_url=None):
     """
 
     """
     ## Determine the parameter, station_id, and dataset_id
     vars1 = list(new_data.variables)
-    dataset_id = [new_data[v].attrs['dataset_id'] for v in vars1 if 'dataset_id' in new_data[v].attrs][0]
+    dataset = [new_data[v].attrs for v in vars1 if 'dataset_id' in new_data[v].attrs][0]
+    dataset_id = dataset['dataset_id']
+    # result_type = dataset['result_type']
     station_id = str(new_data['station_id'].values)
 
     key_dict = {'dataset_id': dataset_id, 'station_id': station_id}
 
+    if read_buffer:
+        base_key_pattern = key_patterns['results_buffer']
+    else:
+        base_key_pattern = key_patterns['results']
+
     ## Get list of keys
-    prefix_key = key_patterns['results'].split('{run_date}')[0].format(**key_dict)
-    all_keys = list_parse_s3(s3, bucket, prefix_key)
-    last_key1 = all_keys[all_keys['KeyDate'] == all_keys['KeyDate'].max()]
+    s3 = s3_connection(conn_config)
+
+    if isinstance(last_run_date_key, str):
+        key_dict.update({'run_date': last_run_date_key})
+        last_key = base_key_pattern.format(**key_dict)
+        try:
+            last_info1 = s3.head_object(Bucket=bucket, Key=last_key)
+            last_key1 = pd.DataFrame({'Key': [last_key]})
+        except:
+            last_key1 = pd.DataFrame()
+    else:
+        prefix_key = key_patterns['results'].split('{run_date}')[0].format(**key_dict)
+        all_keys = list_parse_s3(s3, bucket, prefix_key)
+        last_key1 = all_keys[all_keys['KeyDate'] == all_keys['KeyDate'].max()]
 
     ## Get previous data and compare
+    if isinstance(public_url, str):
+        conn_config = public_url
+
     if not last_key1.empty:
         last_key = last_key1.iloc[0]['Key']
-        obj1 = s3.get_object(Bucket=bucket, Key=last_key)
-        b1 = obj1['Body'].read()
-        p_old_one = read_pkl_zstd(b1, False)
+        # obj1 = s3.get_object(Bucket=bucket, Key=last_key)
+        # b1 = obj1['Body'].read()
+        # p_old_one = read_pkl_zstd(b1, False)
+        p_old_one = get_object_s3(last_key, conn_config, bucket, 'zstd')
         xr_old_one = xr.open_dataset(p_old_one)
         xr_old_one['time'] = xr_old_one['time'].dt.round('s')
 
-        up1 = compare_xrs(xr_old_one, new_data)
+        up1 = compare_xrs(xr_old_one, new_data, add_old=add_old)
     else:
         print('No prior data found in S3. All data will be returned.')
         up1 = new_data.copy()
@@ -1092,46 +1281,235 @@ def compare_datasets_from_s3(s3, bucket, new_data):
 
 
 
-# def update_remote_stations(s3, bucket, dataset_id, station_list, run_date=None):
-#     """
-#
-#     """
-#     if run_date is None:
-#         run_date = pd.Timestamp.today(tz='utc')
-#         run_date_key = run_date.strftime('%Y%m%dT%H%M%SZ')
-#     elif isinstance(run_date, pd.Timestamp):
-#         run_date_key = run_date.strftime('%Y%m%dT%H%M%SZ')
-#     elif isinstance(run_date, str):
-#         run_date_key = run_date
-#     else:
-#         raise TypeError('run_date must be None, Timestamp, or a string representation of a timestamp')
-#
-#     rem_stn = get_remote_station(s3, bucket, dataset_id)
-#
-#     if rem_stn != station_list:
-#         print('stations are different, stations.json.zst will be updated')
-#         all_ids = set([i['station_id'] for i in rem_stn])
-#         all_ids.update(set([i['station_id'] for i in station_list]))
-#
-#         up_list = copy.deepcopy(rem_stn)
-#         up_list = []
-#
-#         for i in all_ids:
-#             rem1 = [n for n in rem_stn if n['station_id'] == i]
-#             new1 = [n for n in station_list if n['station_id'] == i]
-#
-#             if new1:
-#                 up_list.append(new1[0])
-#             else:
-#                 up_list.append(rem1[0])
-#
-#         stn_json1 = write_pkl_zstd(orjson.dumps(up_list))
-#         obj2 = s3.put_object(Bucket=bucket, Key=stn_key, Body=stn_json1, Metadata={'run_date': run_date_key}, ContentType='application/json')
-#     else:
-#         print('stations are the same, stations.json.zst will not be updated')
-#         up_list = rem_stn
-#
-#     return up_list
+def get_filtered_obj_list(remote, dataset_list):
+    """
+
+    """
+    base_prefix = key_patterns['results'].split('{dataset_id}')[0]
+    dataset_ids = [d['dataset_id'] for d in dataset_list]
+
+    s3 = s3_connection(remote['connection_config'])
+    obj_df = list_parse_s3(s3, remote['bucket'], base_prefix)
+
+    obj_df1 = obj_df[obj_df['KeyDate'].notnull()].copy()
+    obj_df1['dataset_id'] = obj_df1['Key'].apply(lambda x: x.split('/')[2])
+    obj_df1['station_id'] = obj_df1['Key'].apply(lambda x: x.split('/')[3])
+    obj_df1['result_type'] = 'results'
+    obj_df1.loc[obj_df1['Key'].str.contains('buffer'), 'result_type'] = 'buffer'
+
+    obj_df2 = obj_df1[obj_df1['dataset_id'].isin(dataset_ids)].copy()
+
+    return obj_df2
+
+
+def get_last_results(obj_df):
+    """
+
+    """
+    last_date1 = obj_df.groupby(['dataset_id', 'station_id'])['KeyDate'].last().reset_index()
+    obj_df1 = pd.merge(last_date1, obj_df[['Key', 'dataset_id', 'station_id', 'KeyDate', 'result_type']], on=['dataset_id', 'station_id', 'KeyDate'])
+
+    return obj_df1
+
+
+def filter_old_ones(obj_df, run_date, days_prior):
+    """
+
+    """
+    if isinstance(run_date.tzname(), str):
+        run_date1 = run_date.tz_convert('UTC').tz_localize(None)
+    else:
+        run_date1 = run_date
+
+    old_date = run_date1 - pd.DateOffset(days=days_prior)
+
+    obj_df1 = obj_df[obj_df['KeyDate'] < old_date].copy()
+
+    return obj_df1
+
+
+def process_buffer(row, remote, run_date_key):
+    """
+
+    """
+    data_list = []
+    for i, r in row.iterrows():
+        obj1 = get_object_s3(r['Key'], remote['connection_config'], remote['bucket'], 'zstd')
+        b1 = io.BytesIO(obj1)
+        xr1 = xr.open_dataset(b1)
+        data_list.append(xr1)
+
+    xr2 = xr.concat(data_list, dim='time', data_vars='minimal')
+    xr3 = xr2.sel(time=~xr2.get_index('time').duplicated('last'))
+    xr3.attrs.update({'history': run_date_key + ': Generated'})
+
+    key_dict = {'dataset_id': r['dataset_id'], 'station_id': r['station_id'], 'run_date': run_date_key}
+    new_key = key_patterns['results'].format(**key_dict)
+
+    cctx = zstd.ZstdCompressor(level=1)
+    c_obj = cctx.compress(xr3.to_netcdf())
+
+    s3 = s3_connection(remote['connection_config'])
+
+    s3.put_object(Body=c_obj, Bucket=remote['bucket'], Key=new_key, ContentType='application/zstd', Metadata={'run_date': run_date_key})
+
+
+def process_buffer_threaded(obj_df, remote, run_date_key, threads=30):
+    """
+
+    """
+    grp1 = obj_df.groupby(['dataset_id', 'station_id'])
+
+    input_list = [[row, remote, run_date_key] for i, row in grp1]
+
+    output = ThreadPool(threads).starmap(process_buffer, input_list)
+
+
+def prepare_station_results(data_dict, dataset_list, station_dict, data_df, run_date_key, mod_date=None, sum_closed='right', other_closed='left', discrete=True):
+    """
+
+    """
+    if isinstance(mod_date, (str, pd.Timestamp)):
+        mod_date = pd.Timestamp(mod_date)
+        ancillary_variables = ['modified_date']
+    else:
+        ancillary_variables = []
+    #     mod_date = pd.Timestamp.today(tz='utc').round('s').tz_localize(None)
+
+    ts_data1 = data_df.copy()
+
+    ## Iterate through each dataset
+    for ds in dataset_list:
+        print(ds['dataset_id'])
+
+        ds_mapping = copy.deepcopy(ds)
+        properties = ds_mapping.pop('properties')
+        attrs = properties['attrs']
+        encoding = properties['encoding']
+
+        attrs1 = copy.deepcopy(attrs)
+        attrs1.update({ds_mapping['parameter']: ds_mapping})
+
+        encoding1 = copy.deepcopy(encoding)
+        encoding1.update({ds_mapping['parameter']: properties['encoding']})
+
+        ## Pre-Process data
+        qual_col = 'quality_code'
+        freq_code = ds_mapping['frequency_interval']
+        parameter = ds_mapping['parameter']
+
+        ## Aggregate data if necessary
+        # Parameter
+        if freq_code == 'T':
+            grp1 = ts_data1.groupby(['time', 'height'])
+            data1 = grp1[parameter].mean()
+
+        else:
+            agg_fun = agg_stat_mapping[ds_mapping['aggregation_statistic']]
+
+            if agg_fun == 'sum':
+                data1 = grp_ts_agg(ts_data1[['time', 'height', parameter]], 'height', 'time', freq_code, agg_fun, closed=sum_closed)
+            else:
+                data1 = grp_ts_agg(ts_data1[['time', 'height', parameter]], 'height', 'time', freq_code, agg_fun, discrete, closed=other_closed)
+
+        # Quality code
+        if qual_col in ts_data1.columns:
+            ts_data1[qual_col] = pd.to_numeric(ts_data1[qual_col], errors='coerce', downcast='integer')
+            if freq_code == 'T':
+                qual1 = grp1[qual_col].min()
+            else:
+                qual1 = grp_ts_agg(ts_data1[['time', 'height', qual_col]], 'height', 'time', freq_code, 'min')
+            df3 = pd.concat([data1, qual1], axis=1).reset_index().dropna()
+
+            ancillary_variables.extend([qual_col])
+        else:
+            df3 = data1.reset_index().copy()
+
+        if 'modified_date' in ancillary_variables:
+            df3['modified_date'] = mod_date
+
+        ## Convert to xarray
+        df4 = df3.copy()
+        df4.set_index(['time', 'height'], inplace=True)
+
+        new1 = data_to_xarray(df4, station_dict, parameter, attrs1, encoding1, run_date=run_date_key, ancillary_variables=ancillary_variables, compression='zstd')
+
+        ## Update the data_dict
+        ds_id = ds_mapping['dataset_id']
+        stn_id = station_dict['station_id']
+
+        data_dict[ds_id].append(new1)
+
+
+def update_results_s3(data_dict, conn_config, bucket, threads=20, add_old=False, read_buffer=False, last_run_date_key=None, public_url=None):
+    """
+
+    """
+    s3 = s3_connection(conn_config, threads)
+
+    for ds_id, results in data_dict.items():
+        print('--dataset_id: ' + ds_id)
+
+        def update_result(result):
+            """
+
+            """
+            ## Process data
+            new1 = xr.open_dataset(read_pkl_zstd(result, False))
+
+            stn_id = str(new1['station_id'].values)
+            print('-station_id: ' + stn_id)
+
+            vars1 = list(new1.variables)
+            parameter = [v for v in vars1 if 'dataset_id' in new1[v].attrs][0]
+            attrs = new1[parameter].attrs.copy()
+
+            ds_id = attrs['dataset_id']
+            run_date_key = new1.attrs['history'].split(':')[0]
+
+            up1 = compare_datasets_from_s3(conn_config, bucket, new1, add_old=add_old, read_buffer=read_buffer, last_run_date_key=last_run_date_key, public_url=public_url)
+
+            ## Save results
+            if isinstance(up1, xr.Dataset) and (len(up1[parameter].time) > 0):
+
+                print('Save results')
+                key_dict = {'dataset_id': ds_id, 'station_id': stn_id, 'run_date': run_date_key}
+
+                new_key = key_patterns['results'].format(**key_dict)
+
+                cctx = zstd.ZstdCompressor(level=1)
+                c_obj = cctx.compress(up1.to_netcdf())
+
+                s3.put_object(Body=c_obj, Bucket=bucket, Key=new_key, ContentType='application/zstd', Metadata={'run_date': run_date_key})
+
+                ## Process stn data
+                print('Save station data')
+
+                stn_m = process_station_summ(ds_id, up1, conn_config, bucket, mod_date=run_date_key)
+
+                stn4 = orjson.loads(stn_m.json(exclude_none=True))
+                up_stns = put_remote_station(s3, bucket, stn4, run_date=run_date_key)
+            else:
+                print('No new data to update')
+
+        ## Run the threadpool
+        output = ThreadPool(threads).imap_unordered(update_result, results)
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
+
 
 #########################################
 ### Testing
